/*
  File autogenerated by gengetopt version 2.22.6
  generated with the following command:
  ggo --no-handle-version --no-handle-error

  The developers of gengetopt consider the fixed text that goes in all
  gengetopt output files to be in the public domain:
  we make no copyright claims on it.
*/

/* If we use autoconf.  */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef FIX_UNUSED
#define FIX_UNUSED(X) (void) (X) /* avoid warnings for unused params */
#endif

#include <getopt.h>

#include "cmdline.h"

const char *gengetopt_args_info_purpose = "SX simulator";

const char *gengetopt_args_info_usage = "Usage: sxsim [OPTIONS]...";

const char *gengetopt_args_info_versiontext = "";

const char *gengetopt_args_info_description = "sxsim can simulate management of an SX cluster. The tool can run in a fully\ninteractive or command-line mode. To enter the interactive mode run it without\nany options or use --execute.";

const char *gengetopt_args_info_help[] = {
  "  -h, --help                    Print help and exit",
  "  -V, --version                 Print version and exit",
  "  -e, --execute=FILE            Execute commands from FILE in interactive mode",
  "\nAll options below are only respected in non-interactive mode.\n\nCluster settings:\n",
  "      --dump-cluster=FILE       Dump content of cluster stored in FILE to\n                                  stdout in CSV format",
  "      --node-list=FILE          Read list of nodes from FILE",
  "      --hash-list=FILE          Read list of hashes from FILE",
  "      --store-data=[N:]FILE/DIR Simulate storing of FILE/DIR and optionally\n                                  force replica count N. This option can be\n                                  used multiple times, also with different\n                                  values for N.",
  "      --replica-count=INT       Set replica count for storage  (default=`1')",
  "\nAutomatic maintainance settings:\n",
  "      --on-lowspace-addnode=NAMEPREFIX@CAPACITY\n                                When cluster gets full add new node(s)\n                                  NAMEPREFIX-ID of size CAPACITY (allows K, M,\n                                  G, T suffixes; default M)",
  "      --on-lowspace-addspace=SIZE\n                                When node gets full resize it by SIZE (allows\n                                  K, M, G, T suffixes; default M)",
  "      --on-upgrade-rebalance    Force rebalance after each upgrade to the\n                                  cluster (instead of automatic mode)\n                                  (default=off)",
  "\nBlock settings:\n",
  "      --blkstats                Print block statistics for input data\n                                  (default=off)",
  "      --block-size=INT          Set block size (in bytes)  (possible\n                                  values=\"4096\", \"8192\", \"16384\",\n                                  \"32768\", \"65536\", \"1048576\",\n                                  \"4194304\" default=`65536')",
  "      --autobs-small=INT        Set small block size for autobs  (possible\n                                  values=\"4096\", \"8192\", \"16384\",\n                                  \"32768\", \"65536\", \"1048576\",\n                                  \"4194304\" default=`4096')",
  "      --autobs-medium=INT       Set medium block size for autobs mode\n                                  (possible values=\"4096\", \"8192\",\n                                  \"16384\", \"32768\", \"65536\", \"1048576\",\n                                  \"4194304\" default=`16384')",
  "      --autobs-big=INT          Set big block size for autobs mode  (possible\n                                  values=\"4096\", \"8192\", \"16384\",\n                                  \"32768\", \"65536\", \"1048576\",\n                                  \"4194304\" default=`1048576')",
  "      --autobs-small-limit=INT  Set maximum file size for small block in autobs\n                                  mode  (default=`131072')",
  "      --autobs-big-limit=INT    Set minimum file size for big block in autobs\n                                  mode  (default=`134217728')",
    0
};

typedef enum {ARG_NO
  , ARG_FLAG
  , ARG_STRING
  , ARG_INT
} cmdline_parser_arg_type;

static
void clear_given (struct gengetopt_args_info *args_info);
static
void clear_args (struct gengetopt_args_info *args_info);

static int
cmdline_parser_internal (int argc, char **argv, struct gengetopt_args_info *args_info,
                        struct cmdline_parser_params *params, const char *additional_error);

static int
cmdline_parser_required2 (struct gengetopt_args_info *args_info, const char *prog_name, const char *additional_error);

const char *cmdline_parser_block_size_values[] = {"4096", "8192", "16384", "32768", "65536", "1048576", "4194304", 0}; /*< Possible values for block-size. */
const char *cmdline_parser_autobs_small_values[] = {"4096", "8192", "16384", "32768", "65536", "1048576", "4194304", 0}; /*< Possible values for autobs-small. */
const char *cmdline_parser_autobs_medium_values[] = {"4096", "8192", "16384", "32768", "65536", "1048576", "4194304", 0}; /*< Possible values for autobs-medium. */
const char *cmdline_parser_autobs_big_values[] = {"4096", "8192", "16384", "32768", "65536", "1048576", "4194304", 0}; /*< Possible values for autobs-big. */

static char *
gengetopt_strdup (const char *s);

static
void clear_given (struct gengetopt_args_info *args_info)
{
  args_info->help_given = 0 ;
  args_info->version_given = 0 ;
  args_info->execute_given = 0 ;
  args_info->dump_cluster_given = 0 ;
  args_info->node_list_given = 0 ;
  args_info->hash_list_given = 0 ;
  args_info->store_data_given = 0 ;
  args_info->replica_count_given = 0 ;
  args_info->on_lowspace_addnode_given = 0 ;
  args_info->on_lowspace_addspace_given = 0 ;
  args_info->on_upgrade_rebalance_given = 0 ;
  args_info->blkstats_given = 0 ;
  args_info->block_size_given = 0 ;
  args_info->autobs_small_given = 0 ;
  args_info->autobs_medium_given = 0 ;
  args_info->autobs_big_given = 0 ;
  args_info->autobs_small_limit_given = 0 ;
  args_info->autobs_big_limit_given = 0 ;
}

static
void clear_args (struct gengetopt_args_info *args_info)
{
  FIX_UNUSED (args_info);
  args_info->execute_arg = NULL;
  args_info->execute_orig = NULL;
  args_info->dump_cluster_arg = NULL;
  args_info->dump_cluster_orig = NULL;
  args_info->node_list_arg = NULL;
  args_info->node_list_orig = NULL;
  args_info->hash_list_arg = NULL;
  args_info->hash_list_orig = NULL;
  args_info->store_data_arg = NULL;
  args_info->store_data_orig = NULL;
  args_info->replica_count_arg = 1;
  args_info->replica_count_orig = NULL;
  args_info->on_lowspace_addnode_arg = NULL;
  args_info->on_lowspace_addnode_orig = NULL;
  args_info->on_lowspace_addspace_arg = NULL;
  args_info->on_lowspace_addspace_orig = NULL;
  args_info->on_upgrade_rebalance_flag = 0;
  args_info->blkstats_flag = 0;
  args_info->block_size_arg = 65536;
  args_info->block_size_orig = NULL;
  args_info->autobs_small_arg = 4096;
  args_info->autobs_small_orig = NULL;
  args_info->autobs_medium_arg = 16384;
  args_info->autobs_medium_orig = NULL;
  args_info->autobs_big_arg = 1048576;
  args_info->autobs_big_orig = NULL;
  args_info->autobs_small_limit_arg = 131072;
  args_info->autobs_small_limit_orig = NULL;
  args_info->autobs_big_limit_arg = 134217728;
  args_info->autobs_big_limit_orig = NULL;
  
}

static
void init_args_info(struct gengetopt_args_info *args_info)
{


  args_info->help_help = gengetopt_args_info_help[0] ;
  args_info->version_help = gengetopt_args_info_help[1] ;
  args_info->execute_help = gengetopt_args_info_help[2] ;
  args_info->dump_cluster_help = gengetopt_args_info_help[4] ;
  args_info->node_list_help = gengetopt_args_info_help[5] ;
  args_info->hash_list_help = gengetopt_args_info_help[6] ;
  args_info->store_data_help = gengetopt_args_info_help[7] ;
  args_info->store_data_min = 0;
  args_info->store_data_max = 0;
  args_info->replica_count_help = gengetopt_args_info_help[8] ;
  args_info->on_lowspace_addnode_help = gengetopt_args_info_help[10] ;
  args_info->on_lowspace_addspace_help = gengetopt_args_info_help[11] ;
  args_info->on_upgrade_rebalance_help = gengetopt_args_info_help[12] ;
  args_info->blkstats_help = gengetopt_args_info_help[14] ;
  args_info->block_size_help = gengetopt_args_info_help[15] ;
  args_info->autobs_small_help = gengetopt_args_info_help[16] ;
  args_info->autobs_medium_help = gengetopt_args_info_help[17] ;
  args_info->autobs_big_help = gengetopt_args_info_help[18] ;
  args_info->autobs_small_limit_help = gengetopt_args_info_help[19] ;
  args_info->autobs_big_limit_help = gengetopt_args_info_help[20] ;
  
}

void
cmdline_parser_print_version (void)
{
  printf ("%s %s\n",
     (strlen(CMDLINE_PARSER_PACKAGE_NAME) ? CMDLINE_PARSER_PACKAGE_NAME : CMDLINE_PARSER_PACKAGE),
     CMDLINE_PARSER_VERSION);

  if (strlen(gengetopt_args_info_versiontext) > 0)
    printf("\n%s\n", gengetopt_args_info_versiontext);
}

static void print_help_common(void) {
  cmdline_parser_print_version ();

  if (strlen(gengetopt_args_info_purpose) > 0)
    printf("\n%s\n", gengetopt_args_info_purpose);

  if (strlen(gengetopt_args_info_usage) > 0)
    printf("\n%s\n", gengetopt_args_info_usage);

  printf("\n");

  if (strlen(gengetopt_args_info_description) > 0)
    printf("%s\n\n", gengetopt_args_info_description);
}

void
cmdline_parser_print_help (void)
{
  int i = 0;
  print_help_common();
  while (gengetopt_args_info_help[i])
    printf("%s\n", gengetopt_args_info_help[i++]);
}

void
cmdline_parser_init (struct gengetopt_args_info *args_info)
{
  clear_given (args_info);
  clear_args (args_info);
  init_args_info (args_info);
}

void
cmdline_parser_params_init(struct cmdline_parser_params *params)
{
  if (params)
    { 
      params->override = 0;
      params->initialize = 1;
      params->check_required = 1;
      params->check_ambiguity = 0;
      params->print_errors = 1;
    }
}

struct cmdline_parser_params *
cmdline_parser_params_create(void)
{
  struct cmdline_parser_params *params = 
    (struct cmdline_parser_params *)malloc(sizeof(struct cmdline_parser_params));
  cmdline_parser_params_init(params);  
  return params;
}

static void
free_string_field (char **s)
{
  if (*s)
    {
      free (*s);
      *s = 0;
    }
}

/** @brief generic value variable */
union generic_value {
    int int_arg;
    char *string_arg;
    const char *default_string_arg;
};

/** @brief holds temporary values for multiple options */
struct generic_list
{
  union generic_value arg;
  char *orig;
  struct generic_list *next;
};

/**
 * @brief add a node at the head of the list 
 */
static void add_node(struct generic_list **list) {
  struct generic_list *new_node = (struct generic_list *) malloc (sizeof (struct generic_list));
  new_node->next = *list;
  *list = new_node;
  new_node->arg.string_arg = 0;
  new_node->orig = 0;
}


static void
free_multiple_string_field(unsigned int len, char ***arg, char ***orig)
{
  unsigned int i;
  if (*arg) {
    for (i = 0; i < len; ++i)
      {
        free_string_field(&((*arg)[i]));
        free_string_field(&((*orig)[i]));
      }
    free_string_field(&((*arg)[0])); /* free default string */

    free (*arg);
    *arg = 0;
    free (*orig);
    *orig = 0;
  }
}

static void
cmdline_parser_release (struct gengetopt_args_info *args_info)
{

  free_string_field (&(args_info->execute_arg));
  free_string_field (&(args_info->execute_orig));
  free_string_field (&(args_info->dump_cluster_arg));
  free_string_field (&(args_info->dump_cluster_orig));
  free_string_field (&(args_info->node_list_arg));
  free_string_field (&(args_info->node_list_orig));
  free_string_field (&(args_info->hash_list_arg));
  free_string_field (&(args_info->hash_list_orig));
  free_multiple_string_field (args_info->store_data_given, &(args_info->store_data_arg), &(args_info->store_data_orig));
  free_string_field (&(args_info->replica_count_orig));
  free_string_field (&(args_info->on_lowspace_addnode_arg));
  free_string_field (&(args_info->on_lowspace_addnode_orig));
  free_string_field (&(args_info->on_lowspace_addspace_arg));
  free_string_field (&(args_info->on_lowspace_addspace_orig));
  free_string_field (&(args_info->block_size_orig));
  free_string_field (&(args_info->autobs_small_orig));
  free_string_field (&(args_info->autobs_medium_orig));
  free_string_field (&(args_info->autobs_big_orig));
  free_string_field (&(args_info->autobs_small_limit_orig));
  free_string_field (&(args_info->autobs_big_limit_orig));
  
  

  clear_given (args_info);
}

/**
 * @param val the value to check
 * @param values the possible values
 * @return the index of the matched value:
 * -1 if no value matched,
 * -2 if more than one value has matched
 */
static int
check_possible_values(const char *val, const char *values[])
{
  int i, found, last;
  size_t len;

  if (!val)   /* otherwise strlen() crashes below */
    return -1; /* -1 means no argument for the option */

  found = last = 0;

  for (i = 0, len = strlen(val); values[i]; ++i)
    {
      if (strncmp(val, values[i], len) == 0)
        {
          ++found;
          last = i;
          if (strlen(values[i]) == len)
            return i; /* exact macth no need to check more */
        }
    }

  if (found == 1) /* one match: OK */
    return last;

  return (found ? -2 : -1); /* return many values or none matched */
}


static void
write_into_file(FILE *outfile, const char *opt, const char *arg, const char *values[])
{
  int found = -1;
  if (arg) {
    if (values) {
      found = check_possible_values(arg, values);      
    }
    if (found >= 0)
      fprintf(outfile, "%s=\"%s\" # %s\n", opt, arg, values[found]);
    else
      fprintf(outfile, "%s=\"%s\"\n", opt, arg);
  } else {
    fprintf(outfile, "%s\n", opt);
  }
}

static void
write_multiple_into_file(FILE *outfile, int len, const char *opt, char **arg, const char *values[])
{
  int i;
  
  for (i = 0; i < len; ++i)
    write_into_file(outfile, opt, (arg ? arg[i] : 0), values);
}

int
cmdline_parser_dump(FILE *outfile, struct gengetopt_args_info *args_info)
{
  int i = 0;

  if (!outfile)
    {
      fprintf (stderr, "%s: cannot dump options to stream\n", CMDLINE_PARSER_PACKAGE);
      return EXIT_FAILURE;
    }

  if (args_info->help_given)
    write_into_file(outfile, "help", 0, 0 );
  if (args_info->version_given)
    write_into_file(outfile, "version", 0, 0 );
  if (args_info->execute_given)
    write_into_file(outfile, "execute", args_info->execute_orig, 0);
  if (args_info->dump_cluster_given)
    write_into_file(outfile, "dump-cluster", args_info->dump_cluster_orig, 0);
  if (args_info->node_list_given)
    write_into_file(outfile, "node-list", args_info->node_list_orig, 0);
  if (args_info->hash_list_given)
    write_into_file(outfile, "hash-list", args_info->hash_list_orig, 0);
  write_multiple_into_file(outfile, args_info->store_data_given, "store-data", args_info->store_data_orig, 0);
  if (args_info->replica_count_given)
    write_into_file(outfile, "replica-count", args_info->replica_count_orig, 0);
  if (args_info->on_lowspace_addnode_given)
    write_into_file(outfile, "on-lowspace-addnode", args_info->on_lowspace_addnode_orig, 0);
  if (args_info->on_lowspace_addspace_given)
    write_into_file(outfile, "on-lowspace-addspace", args_info->on_lowspace_addspace_orig, 0);
  if (args_info->on_upgrade_rebalance_given)
    write_into_file(outfile, "on-upgrade-rebalance", 0, 0 );
  if (args_info->blkstats_given)
    write_into_file(outfile, "blkstats", 0, 0 );
  if (args_info->block_size_given)
    write_into_file(outfile, "block-size", args_info->block_size_orig, cmdline_parser_block_size_values);
  if (args_info->autobs_small_given)
    write_into_file(outfile, "autobs-small", args_info->autobs_small_orig, cmdline_parser_autobs_small_values);
  if (args_info->autobs_medium_given)
    write_into_file(outfile, "autobs-medium", args_info->autobs_medium_orig, cmdline_parser_autobs_medium_values);
  if (args_info->autobs_big_given)
    write_into_file(outfile, "autobs-big", args_info->autobs_big_orig, cmdline_parser_autobs_big_values);
  if (args_info->autobs_small_limit_given)
    write_into_file(outfile, "autobs-small-limit", args_info->autobs_small_limit_orig, 0);
  if (args_info->autobs_big_limit_given)
    write_into_file(outfile, "autobs-big-limit", args_info->autobs_big_limit_orig, 0);
  

  i = EXIT_SUCCESS;
  return i;
}

int
cmdline_parser_file_save(const char *filename, struct gengetopt_args_info *args_info)
{
  FILE *outfile;
  int i = 0;

  outfile = fopen(filename, "w");

  if (!outfile)
    {
      fprintf (stderr, "%s: cannot open file for writing: %s\n", CMDLINE_PARSER_PACKAGE, filename);
      return EXIT_FAILURE;
    }

  i = cmdline_parser_dump(outfile, args_info);
  fclose (outfile);

  return i;
}

void
cmdline_parser_free (struct gengetopt_args_info *args_info)
{
  cmdline_parser_release (args_info);
}

/** @brief replacement of strdup, which is not standard */
char *
gengetopt_strdup (const char *s)
{
  char *result = 0;
  if (!s)
    return result;

  result = (char*)malloc(strlen(s) + 1);
  if (result == (char*)0)
    return (char*)0;
  strcpy(result, s);
  return result;
}

static char *
get_multiple_arg_token(const char *arg)
{
  const char *tok;
  char *ret;
  size_t len, num_of_escape, i, j;

  if (!arg)
    return 0;

  tok = strchr (arg, ',');
  num_of_escape = 0;

  /* make sure it is not escaped */
  while (tok)
    {
      if (*(tok-1) == '\\')
        {
          /* find the next one */
          tok = strchr (tok+1, ',');
          ++num_of_escape;
        }
      else
        break;
    }

  if (tok)
    len = (size_t)(tok - arg + 1);
  else
    len = strlen (arg) + 1;

  len -= num_of_escape;

  ret = (char *) malloc (len);

  i = 0;
  j = 0;
  while (arg[i] && (j < len-1))
    {
      if (arg[i] == '\\' && 
	  arg[ i + 1 ] && 
	  arg[ i + 1 ] == ',')
        ++i;

      ret[j++] = arg[i++];
    }

  ret[len-1] = '\0';

  return ret;
}

static const char *
get_multiple_arg_token_next(const char *arg)
{
  const char *tok;

  if (!arg)
    return 0;

  tok = strchr (arg, ',');

  /* make sure it is not escaped */
  while (tok)
    {
      if (*(tok-1) == '\\')
        {
          /* find the next one */
          tok = strchr (tok+1, ',');
        }
      else
        break;
    }

  if (! tok || strlen(tok) == 1)
    return 0;

  return tok+1;
}

static int
check_multiple_option_occurrences(const char *prog_name, unsigned int option_given, unsigned int min, unsigned int max, const char *option_desc);

int
check_multiple_option_occurrences(const char *prog_name, unsigned int option_given, unsigned int min, unsigned int max, const char *option_desc)
{
  int error_occurred = 0;

  if (option_given && (min > 0 || max > 0))
    {
      if (min > 0 && max > 0)
        {
          if (min == max)
            {
              /* specific occurrences */
              if (option_given != (unsigned int) min)
                {
                  fprintf (stderr, "%s: %s option occurrences must be %d\n",
                    prog_name, option_desc, min);
                  error_occurred = 1;
                }
            }
          else if (option_given < (unsigned int) min
                || option_given > (unsigned int) max)
            {
              /* range occurrences */
              fprintf (stderr, "%s: %s option occurrences must be between %d and %d\n",
                prog_name, option_desc, min, max);
              error_occurred = 1;
            }
        }
      else if (min > 0)
        {
          /* at least check */
          if (option_given < min)
            {
              fprintf (stderr, "%s: %s option occurrences must be at least %d\n",
                prog_name, option_desc, min);
              error_occurred = 1;
            }
        }
      else if (max > 0)
        {
          /* at most check */
          if (option_given > max)
            {
              fprintf (stderr, "%s: %s option occurrences must be at most %d\n",
                prog_name, option_desc, max);
              error_occurred = 1;
            }
        }
    }
    
  return error_occurred;
}
int
cmdline_parser (int argc, char **argv, struct gengetopt_args_info *args_info)
{
  return cmdline_parser2 (argc, argv, args_info, 0, 1, 1);
}

int
cmdline_parser_ext (int argc, char **argv, struct gengetopt_args_info *args_info,
                   struct cmdline_parser_params *params)
{
  int result;
  result = cmdline_parser_internal (argc, argv, args_info, params, 0);

  return result;
}

int
cmdline_parser2 (int argc, char **argv, struct gengetopt_args_info *args_info, int override, int initialize, int check_required)
{
  int result;
  struct cmdline_parser_params params;
  
  params.override = override;
  params.initialize = initialize;
  params.check_required = check_required;
  params.check_ambiguity = 0;
  params.print_errors = 1;

  result = cmdline_parser_internal (argc, argv, args_info, &params, 0);

  return result;
}

int
cmdline_parser_required (struct gengetopt_args_info *args_info, const char *prog_name)
{
  int result = EXIT_SUCCESS;

  if (cmdline_parser_required2(args_info, prog_name, 0) > 0)
    result = EXIT_FAILURE;

  return result;
}

int
cmdline_parser_required2 (struct gengetopt_args_info *args_info, const char *prog_name, const char *additional_error)
{
  int error_occurred = 0;
  FIX_UNUSED (additional_error);

  /* checks for required options */
  if (check_multiple_option_occurrences(prog_name, args_info->store_data_given, args_info->store_data_min, args_info->store_data_max, "'--store-data'"))
     error_occurred = 1;
  
  
  /* checks for dependences among options */

  return error_occurred;
}


static char *package_name = 0;

/**
 * @brief updates an option
 * @param field the generic pointer to the field to update
 * @param orig_field the pointer to the orig field
 * @param field_given the pointer to the number of occurrence of this option
 * @param prev_given the pointer to the number of occurrence already seen
 * @param value the argument for this option (if null no arg was specified)
 * @param possible_values the possible values for this option (if specified)
 * @param default_value the default value (in case the option only accepts fixed values)
 * @param arg_type the type of this option
 * @param check_ambiguity @see cmdline_parser_params.check_ambiguity
 * @param override @see cmdline_parser_params.override
 * @param no_free whether to free a possible previous value
 * @param multiple_option whether this is a multiple option
 * @param long_opt the corresponding long option
 * @param short_opt the corresponding short option (or '-' if none)
 * @param additional_error possible further error specification
 */
static
int update_arg(void *field, char **orig_field,
               unsigned int *field_given, unsigned int *prev_given, 
               char *value, const char *possible_values[],
               const char *default_value,
               cmdline_parser_arg_type arg_type,
               int check_ambiguity, int override,
               int no_free, int multiple_option,
               const char *long_opt, char short_opt,
               const char *additional_error)
{
  char *stop_char = 0;
  const char *val = value;
  int found;
  char **string_field;
  FIX_UNUSED (field);

  stop_char = 0;
  found = 0;

  if (!multiple_option && prev_given && (*prev_given || (check_ambiguity && *field_given)))
    {
      if (short_opt != '-')
        fprintf (stderr, "%s: `--%s' (`-%c') option given more than once%s\n", 
               package_name, long_opt, short_opt,
               (additional_error ? additional_error : ""));
      else
        fprintf (stderr, "%s: `--%s' option given more than once%s\n", 
               package_name, long_opt,
               (additional_error ? additional_error : ""));
      return 1; /* failure */
    }

  if (possible_values && (found = check_possible_values((value ? value : default_value), possible_values)) < 0)
    {
      if (short_opt != '-')
        fprintf (stderr, "%s: %s argument, \"%s\", for option `--%s' (`-%c')%s\n", 
          package_name, (found == -2) ? "ambiguous" : "invalid", value, long_opt, short_opt,
          (additional_error ? additional_error : ""));
      else
        fprintf (stderr, "%s: %s argument, \"%s\", for option `--%s'%s\n", 
          package_name, (found == -2) ? "ambiguous" : "invalid", value, long_opt,
          (additional_error ? additional_error : ""));
      return 1; /* failure */
    }
    
  if (field_given && *field_given && ! override)
    return 0;
  if (prev_given)
    (*prev_given)++;
  if (field_given)
    (*field_given)++;
  if (possible_values)
    val = possible_values[found];

  switch(arg_type) {
  case ARG_FLAG:
    *((int *)field) = !*((int *)field);
    break;
  case ARG_INT:
    if (val) *((int *)field) = strtol (val, &stop_char, 0);
    break;
  case ARG_STRING:
    if (val) {
      string_field = (char **)field;
      if (!no_free && *string_field)
        free (*string_field); /* free previous string */
      *string_field = gengetopt_strdup (val);
    }
    break;
  default:
    break;
  };

  /* check numeric conversion */
  switch(arg_type) {
  case ARG_INT:
    if (val && !(stop_char && *stop_char == '\0')) {
      fprintf(stderr, "%s: invalid numeric value: %s\n", package_name, val);
      return 1; /* failure */
    }
    break;
  default:
    ;
  };

  /* store the original value */
  switch(arg_type) {
  case ARG_NO:
  case ARG_FLAG:
    break;
  default:
    if (value && orig_field) {
      if (no_free) {
        *orig_field = value;
      } else {
        if (*orig_field)
          free (*orig_field); /* free previous string */
        *orig_field = gengetopt_strdup (value);
      }
    }
  };

  return 0; /* OK */
}

/**
 * @brief store information about a multiple option in a temporary list
 * @param list where to (temporarily) store multiple options
 */
static
int update_multiple_arg_temp(struct generic_list **list,
               unsigned int *prev_given, const char *val,
               const char *possible_values[], const char *default_value,
               cmdline_parser_arg_type arg_type,
               const char *long_opt, char short_opt,
               const char *additional_error)
{
  /* store single arguments */
  char *multi_token;
  const char *multi_next;

  if (arg_type == ARG_NO) {
    (*prev_given)++;
    return 0; /* OK */
  }

  multi_token = get_multiple_arg_token(val);
  multi_next = get_multiple_arg_token_next (val);

  while (1)
    {
      add_node (list);
      if (update_arg((void *)&((*list)->arg), &((*list)->orig), 0,
          prev_given, multi_token, possible_values, default_value, 
          arg_type, 0, 1, 1, 1, long_opt, short_opt, additional_error)) {
        if (multi_token) free(multi_token);
        return 1; /* failure */
      }

      if (multi_next)
        {
          multi_token = get_multiple_arg_token(multi_next);
          multi_next = get_multiple_arg_token_next (multi_next);
        }
      else
        break;
    }

  return 0; /* OK */
}

/**
 * @brief free the passed list (including possible string argument)
 */
static
void free_list(struct generic_list *list, short string_arg)
{
  if (list) {
    struct generic_list *tmp;
    while (list)
      {
        tmp = list;
        if (string_arg && list->arg.string_arg)
          free (list->arg.string_arg);
        if (list->orig)
          free (list->orig);
        list = list->next;
        free (tmp);
      }
  }
}

/**
 * @brief updates a multiple option starting from the passed list
 */
static
void update_multiple_arg(void *field, char ***orig_field,
               unsigned int field_given, unsigned int prev_given, union generic_value *default_value,
               cmdline_parser_arg_type arg_type,
               struct generic_list *list)
{
  int i;
  struct generic_list *tmp;

  if (prev_given && list) {
    *orig_field = (char **) realloc (*orig_field, (field_given + prev_given) * sizeof (char *));

    switch(arg_type) {
    case ARG_INT:
      *((int **)field) = (int *)realloc (*((int **)field), (field_given + prev_given) * sizeof (int)); break;
    case ARG_STRING:
      *((char ***)field) = (char **)realloc (*((char ***)field), (field_given + prev_given) * sizeof (char *)); break;
    default:
      break;
    };
    
    for (i = (prev_given - 1); i >= 0; --i)
      {
        tmp = list;
        
        switch(arg_type) {
        case ARG_INT:
          (*((int **)field))[i + field_given] = tmp->arg.int_arg; break;
        case ARG_STRING:
          (*((char ***)field))[i + field_given] = tmp->arg.string_arg; break;
        default:
          break;
        }        
        (*orig_field) [i + field_given] = list->orig;
        list = list->next;
        free (tmp);
      }
  } else { /* set the default value */
    if (default_value && ! field_given) {
      switch(arg_type) {
      case ARG_INT:
        if (! *((int **)field)) {
          *((int **)field) = (int *)malloc (sizeof (int));
          (*((int **)field))[0] = default_value->int_arg; 
        }
        break;
      case ARG_STRING:
        if (! *((char ***)field)) {
          *((char ***)field) = (char **)malloc (sizeof (char *));
          (*((char ***)field))[0] = gengetopt_strdup(default_value->string_arg);
        }
        break;
      default: break;
      }
      if (!(*orig_field)) {
        *orig_field = (char **) malloc (sizeof (char *));
        (*orig_field)[0] = 0;
      }
    }
  }
}

int
cmdline_parser_internal (
  int argc, char **argv, struct gengetopt_args_info *args_info,
                        struct cmdline_parser_params *params, const char *additional_error)
{
  int c;	/* Character of the parsed option.  */

  struct generic_list * store_data_list = NULL;
  int error_occurred = 0;
  struct gengetopt_args_info local_args_info;
  
  int override;
  int initialize;
  int check_required;
  int check_ambiguity;
  
  package_name = argv[0];
  
  override = params->override;
  initialize = params->initialize;
  check_required = params->check_required;
  check_ambiguity = params->check_ambiguity;

  if (initialize)
    cmdline_parser_init (args_info);

  cmdline_parser_init (&local_args_info);

  optarg = 0;
  optind = 0;
  opterr = params->print_errors;
  optopt = '?';

  while (1)
    {
      int option_index = 0;

      static struct option long_options[] = {
        { "help",	0, NULL, 'h' },
        { "version",	0, NULL, 'V' },
        { "execute",	1, NULL, 'e' },
        { "dump-cluster",	1, NULL, 0 },
        { "node-list",	1, NULL, 0 },
        { "hash-list",	1, NULL, 0 },
        { "store-data",	1, NULL, 0 },
        { "replica-count",	1, NULL, 0 },
        { "on-lowspace-addnode",	1, NULL, 0 },
        { "on-lowspace-addspace",	1, NULL, 0 },
        { "on-upgrade-rebalance",	0, NULL, 0 },
        { "blkstats",	0, NULL, 0 },
        { "block-size",	1, NULL, 0 },
        { "autobs-small",	1, NULL, 0 },
        { "autobs-medium",	1, NULL, 0 },
        { "autobs-big",	1, NULL, 0 },
        { "autobs-small-limit",	1, NULL, 0 },
        { "autobs-big-limit",	1, NULL, 0 },
        { 0,  0, 0, 0 }
      };

      c = getopt_long (argc, argv, "hVe:", long_options, &option_index);

      if (c == -1) break;	/* Exit from `while (1)' loop.  */

      switch (c)
        {
        case 'h':	/* Print help and exit.  */
          cmdline_parser_print_help ();
          cmdline_parser_free (&local_args_info);
          exit (EXIT_SUCCESS);

        case 'V':	/* Print version and exit.  */
        
        
          if (update_arg( 0 , 
               0 , &(args_info->version_given),
              &(local_args_info.version_given), optarg, 0, 0, ARG_NO,
              check_ambiguity, override, 0, 0,
              "version", 'V',
              additional_error))
            goto failure;
          cmdline_parser_free (&local_args_info);
          return 0;
        
          break;
        case 'e':	/* Execute commands from FILE in interactive mode.  */
        
        
          if (update_arg( (void *)&(args_info->execute_arg), 
               &(args_info->execute_orig), &(args_info->execute_given),
              &(local_args_info.execute_given), optarg, 0, 0, ARG_STRING,
              check_ambiguity, override, 0, 0,
              "execute", 'e',
              additional_error))
            goto failure;
        
          break;

        case 0:	/* Long option with no short option */
          /* Dump content of cluster stored in FILE to stdout in CSV format.  */
          if (strcmp (long_options[option_index].name, "dump-cluster") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->dump_cluster_arg), 
                 &(args_info->dump_cluster_orig), &(args_info->dump_cluster_given),
                &(local_args_info.dump_cluster_given), optarg, 0, 0, ARG_STRING,
                check_ambiguity, override, 0, 0,
                "dump-cluster", '-',
                additional_error))
              goto failure;
          
          }
          /* Read list of nodes from FILE.  */
          else if (strcmp (long_options[option_index].name, "node-list") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->node_list_arg), 
                 &(args_info->node_list_orig), &(args_info->node_list_given),
                &(local_args_info.node_list_given), optarg, 0, 0, ARG_STRING,
                check_ambiguity, override, 0, 0,
                "node-list", '-',
                additional_error))
              goto failure;
          
          }
          /* Read list of hashes from FILE.  */
          else if (strcmp (long_options[option_index].name, "hash-list") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->hash_list_arg), 
                 &(args_info->hash_list_orig), &(args_info->hash_list_given),
                &(local_args_info.hash_list_given), optarg, 0, 0, ARG_STRING,
                check_ambiguity, override, 0, 0,
                "hash-list", '-',
                additional_error))
              goto failure;
          
          }
          /* Simulate storing of FILE/DIR and optionally force replica count N. This option can be used multiple times, also with different values for N..  */
          else if (strcmp (long_options[option_index].name, "store-data") == 0)
          {
          
            if (update_multiple_arg_temp(&store_data_list, 
                &(local_args_info.store_data_given), optarg, 0, 0, ARG_STRING,
                "store-data", '-',
                additional_error))
              goto failure;
          
          }
          /* Set replica count for storage.  */
          else if (strcmp (long_options[option_index].name, "replica-count") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->replica_count_arg), 
                 &(args_info->replica_count_orig), &(args_info->replica_count_given),
                &(local_args_info.replica_count_given), optarg, 0, "1", ARG_INT,
                check_ambiguity, override, 0, 0,
                "replica-count", '-',
                additional_error))
              goto failure;
          
          }
          /* When cluster gets full add new node(s) NAMEPREFIX-ID of size CAPACITY (allows K, M, G, T suffixes; default M).  */
          else if (strcmp (long_options[option_index].name, "on-lowspace-addnode") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->on_lowspace_addnode_arg), 
                 &(args_info->on_lowspace_addnode_orig), &(args_info->on_lowspace_addnode_given),
                &(local_args_info.on_lowspace_addnode_given), optarg, 0, 0, ARG_STRING,
                check_ambiguity, override, 0, 0,
                "on-lowspace-addnode", '-',
                additional_error))
              goto failure;
          
          }
          /* When node gets full resize it by SIZE (allows K, M, G, T suffixes; default M).  */
          else if (strcmp (long_options[option_index].name, "on-lowspace-addspace") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->on_lowspace_addspace_arg), 
                 &(args_info->on_lowspace_addspace_orig), &(args_info->on_lowspace_addspace_given),
                &(local_args_info.on_lowspace_addspace_given), optarg, 0, 0, ARG_STRING,
                check_ambiguity, override, 0, 0,
                "on-lowspace-addspace", '-',
                additional_error))
              goto failure;
          
          }
          /* Force rebalance after each upgrade to the cluster (instead of automatic mode).  */
          else if (strcmp (long_options[option_index].name, "on-upgrade-rebalance") == 0)
          {
          
          
            if (update_arg((void *)&(args_info->on_upgrade_rebalance_flag), 0, &(args_info->on_upgrade_rebalance_given),
                &(local_args_info.on_upgrade_rebalance_given), optarg, 0, 0, ARG_FLAG,
                check_ambiguity, override, 1, 0, "on-upgrade-rebalance", '-',
                additional_error))
              goto failure;
          
          }
          /* Print block statistics for input data.  */
          else if (strcmp (long_options[option_index].name, "blkstats") == 0)
          {
          
          
            if (update_arg((void *)&(args_info->blkstats_flag), 0, &(args_info->blkstats_given),
                &(local_args_info.blkstats_given), optarg, 0, 0, ARG_FLAG,
                check_ambiguity, override, 1, 0, "blkstats", '-',
                additional_error))
              goto failure;
          
          }
          /* Set block size (in bytes).  */
          else if (strcmp (long_options[option_index].name, "block-size") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->block_size_arg), 
                 &(args_info->block_size_orig), &(args_info->block_size_given),
                &(local_args_info.block_size_given), optarg, cmdline_parser_block_size_values, "65536", ARG_INT,
                check_ambiguity, override, 0, 0,
                "block-size", '-',
                additional_error))
              goto failure;
          
          }
          /* Set small block size for autobs.  */
          else if (strcmp (long_options[option_index].name, "autobs-small") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->autobs_small_arg), 
                 &(args_info->autobs_small_orig), &(args_info->autobs_small_given),
                &(local_args_info.autobs_small_given), optarg, cmdline_parser_autobs_small_values, "4096", ARG_INT,
                check_ambiguity, override, 0, 0,
                "autobs-small", '-',
                additional_error))
              goto failure;
          
          }
          /* Set medium block size for autobs mode.  */
          else if (strcmp (long_options[option_index].name, "autobs-medium") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->autobs_medium_arg), 
                 &(args_info->autobs_medium_orig), &(args_info->autobs_medium_given),
                &(local_args_info.autobs_medium_given), optarg, cmdline_parser_autobs_medium_values, "16384", ARG_INT,
                check_ambiguity, override, 0, 0,
                "autobs-medium", '-',
                additional_error))
              goto failure;
          
          }
          /* Set big block size for autobs mode.  */
          else if (strcmp (long_options[option_index].name, "autobs-big") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->autobs_big_arg), 
                 &(args_info->autobs_big_orig), &(args_info->autobs_big_given),
                &(local_args_info.autobs_big_given), optarg, cmdline_parser_autobs_big_values, "1048576", ARG_INT,
                check_ambiguity, override, 0, 0,
                "autobs-big", '-',
                additional_error))
              goto failure;
          
          }
          /* Set maximum file size for small block in autobs mode.  */
          else if (strcmp (long_options[option_index].name, "autobs-small-limit") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->autobs_small_limit_arg), 
                 &(args_info->autobs_small_limit_orig), &(args_info->autobs_small_limit_given),
                &(local_args_info.autobs_small_limit_given), optarg, 0, "131072", ARG_INT,
                check_ambiguity, override, 0, 0,
                "autobs-small-limit", '-',
                additional_error))
              goto failure;
          
          }
          /* Set minimum file size for big block in autobs mode.  */
          else if (strcmp (long_options[option_index].name, "autobs-big-limit") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->autobs_big_limit_arg), 
                 &(args_info->autobs_big_limit_orig), &(args_info->autobs_big_limit_given),
                &(local_args_info.autobs_big_limit_given), optarg, 0, "134217728", ARG_INT,
                check_ambiguity, override, 0, 0,
                "autobs-big-limit", '-',
                additional_error))
              goto failure;
          
          }
          
          break;
        case '?':	/* Invalid option.  */
          /* `getopt_long' already printed an error message.  */
          goto failure;

        default:	/* bug: option not considered.  */
          fprintf (stderr, "%s: option unknown: %c%s\n", CMDLINE_PARSER_PACKAGE, c, (additional_error ? additional_error : ""));
          abort ();
        } /* switch */
    } /* while */


  update_multiple_arg((void *)&(args_info->store_data_arg),
    &(args_info->store_data_orig), args_info->store_data_given,
    local_args_info.store_data_given, 0,
    ARG_STRING, store_data_list);

  args_info->store_data_given += local_args_info.store_data_given;
  local_args_info.store_data_given = 0;
  
  if (check_required)
    {
      error_occurred += cmdline_parser_required2 (args_info, argv[0], additional_error);
    }

  cmdline_parser_release (&local_args_info);

  if ( error_occurred )
    return (EXIT_FAILURE);

  return 0;

failure:
  free_list (store_data_list, 1 );
  
  cmdline_parser_release (&local_args_info);
  return (EXIT_FAILURE);
}
